import unittest
import os

from test_base_connection import TestServerClient
from ai_server import VectorEngine

from variables import VECTOR_ENGINE_ID

class VectorTests(TestServerClient):
    
    vector_engine = None
    test_files = None
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.vector_engine = VectorEngine(
            engine_id = VECTOR_ENGINE_ID, 
        )
        
        if cls.test_files is None:
            CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
            TEST_FILE_DIR = os.path.join(CURRENT_DIR, "test_files")
            # List files with full paths
            TEST_FILES = [os.path.join(TEST_FILE_DIR, file) for file in os.listdir(TEST_FILE_DIR)]
            cls.test_files = TEST_FILES
        
    def test_vector_search(self):
        from pathlib import Path
        
        self.vector_engine.addDocument(
            file_paths=self.test_files
        )

        file_names = [Path(file).name for file in self.test_files]
    
        vector_search = self.vector_engine.nearestNeighbor(
            search_statement = 'How did the WHO improve access to oxygen supplies?', 
            limit = 5
        )
        
        # Assert that the response is a dictionary with specific keys
        self.assertIsInstance(vector_search, list)
        
        top_match = vector_search[0]
        self.assertIsInstance(top_match, dict)
        
        self.assertCountEqual(top_match.keys(), ['Score', 'Source', 'Modality' , 'Divider', 'Part', 'Tokens', 'Content'])

        self.vector_engine.removeDocument(file_names=file_names)
        
    def test_list_documents(self):

        document_info_list = self.vector_engine.listDocuments()[0]

        self.assertIsInstance(document_info_list, list)
        
        if len(document_info_list) > 0:   
            document_info = document_info_list[0]
            self.assertIsInstance(document_info, dict)
            self.assertIn('fileName', document_info)

if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd

from ai_server import DatabaseEngine

from test_base_connection import TestServerClient
from variables import DATABASE_ENGINE_ID

class DatabaseTests(TestServerClient):
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.db_engine = DatabaseEngine(
            engine_id = DATABASE_ENGINE_ID, 
        )  
        
    def test_db_query(self):
        db_pandas = self.db_engine.execQuery('select * from diabetes')
        
        # Assert that the response is a dictionary with specific keys
        self.assertIsInstance(db_pandas, pd.DataFrame)

if __name__ == '__main__':
    unittest.main()



import unittest
from test_base_connection import TestServerClient

class TestRunPixel(TestServerClient):
    
    def test_server_connection(self):
        # Mock response from the server
        expected_response = 2

        # Test the server connection
        response = self.server_client.run_pixel('1+1')  # Replace some_method with the actual method you're testing
        self.assertEqual(response, expected_response)

if __name__ == '__main__':
    unittest.main()



from setuptools import find_packages, setup

with open('README.md','r') as f:
    long_description = f.read()
    
setup(
    name="ai-server-sdk",                     # This is the name of the package
    version="0.0.17",
    packages=find_packages(),
    install_requires=[
        'requests', 
        'pandas', 
        'jsonpickle',
    ], 
    extras_require={
        "full": ["langchain", "langchain-community"]
    },
    author="Thomas Trankle, Maher Khalil, Ryan Weiler",
    description="Utility package to connect to AI Server instances.",
    license="MIT",
    long_description = long_description,
    long_description_content_type = 'text/markdown'
)


import unittest
from ai_server import ServerClient
from variables import ACCESS_KEY, SECRET_KEY, ENDPOINT

class TestServerClient(unittest.TestCase):
    
    server_client = None
    
    @classmethod
    def setUpClass(cls):
        # Create an instance of ServerClient for testing
        if cls.server_client is None:
            cls.server_client = TestServerClient.login_with_access_keys()
            
            
    @staticmethod
    def login_with_access_keys():
        return ServerClient(
            access_key=ACCESS_KEY,
            secret_key=SECRET_KEY,
            base=ENDPOINT
        )

if __name__ == '__main__':
    unittest.main()



import unittest
from typing import Generator
from ai_server import ModelEngine

from test_base_connection import TestServerClient
from variables import LLM_EMBEDDING_ENGINE_ID, LLM_CHAT_ENGINE_ID

class ModelTests(TestServerClient):
    
    def test_model_ask(self):
        model = ModelEngine(
            engine_id = LLM_CHAT_ENGINE_ID, 
        )

        model_response_list = model.ask(question = 'what is the capital of france ey?')

        # Assert that the response is a dictionary with specific keys
        self.assertIsInstance(model_response_list, list)
        
        model_response = model_response_list[0]
        self.assertIsInstance(model_response, dict)
        
        self.assertCountEqual(model_response.keys(), ['response', 'numberOfTokensInPrompt', 'numberOfTokensInResponse' , 'messageId', 'roomId'])
        
    def test_model_ask_with_stream(self):
        
        model = ModelEngine(
            engine_id = LLM_CHAT_ENGINE_ID, 
        )

        model_response_stream = model.stream_ask(question = 'what is the capital of france ey?')

        self.assertIsInstance(model_response_stream, Generator)
        
    def test_model_embeddings(self):
        model = ModelEngine(
            engine_id = LLM_EMBEDDING_ENGINE_ID, 
        )

        model_response_list = model.embeddings(strings_to_embed = ['what is the capital of france ey?'])

        # Assert that the response is a dictionary with specific keys
        self.assertIsInstance(model_response_list, list)
        
        model_response = model_response_list[0]
        self.assertIsInstance(model_response, dict)
        
        self.assertCountEqual(model_response.keys(), ['response', 'numberOfTokensInPrompt', 'numberOfTokensInResponse'])
        

if __name__ == '__main__':
    #python test_model.py -v -m "access_key=my_access_key secret_key=my_secret_key base=http://example.com/api"

    unittest.main()



from typing import (
    List,
    Dict,
    Optional
)
import json
import logging
from ...server_resources.server_proxy import ServerProxy

logger: logging.Logger = logging.getLogger(__name__)


class VectorEngine(ServerProxy):
    '''Python class to interact with Vector Database Engines defined in CFG AI'''

    engine_type = 'VECTOR'

    def __init__(
        self,
        engine_id: str,
        insight_id: Optional[str] = None,
    ):
        super().__init__()
        self.engine_id = engine_id
        self.insight_id = insight_id

        logger.info("VectorEngine initialized with engine id " + engine_id)

    def addDocument(
        self,
        file_paths: List[str],
        param_dict: Optional[Dict] = {},
        insight_id: Optional[str] = None,
    ) -> None:
        '''
        This method is used to add documents to a vector database. The engine itself will determine how the the documents are 
        processed and the embeddings are created.

        Args:
            file_paths (`List[str]`): List of local files paths to push to the server and index.
            param_dict (`Optional[Dict]`): Additional parameters the engine might need to process the documents.
            insight_id (`Optional[str]`): The insight ID to upload the documents to and process the request. Default is to use the clients current insight.

        Returns:
            `None`
        '''
        if insight_id is None:
            if self.insight_id is None:
                insight_id = self.insight_id
            else:
                insight_id = self.server.cur_insight

        assert self.server != None
        insight_files = self.server.upload_files(
            files=file_paths,
            insight_id=insight_id,
        )

        pixel = "CreateEmbeddingsFromDocuments (engine = \"" + \
            self.engine_id + "\", filePaths = "+json.dumps(insight_files)

        if len(param_dict) != 0:
            pixel += ", paramValues = " + json.dumps(param_dict)

        pixel += ");"

        output_payload_message = self.server.run_pixel(
            payload=pixel,
            insight_id=insight_id,
            full_response=True
        )

        if (output_payload_message['pixelReturn'][0]['operationType'] == ['ERROR']):
            raise RuntimeError(
                output_payload_message['pixelReturn'][0]['output'])

        # output_payload_message['pixelReturn'][0]['output']

    def removeDocument(
        self,
        file_names: List[str],
        param_dict: Optional[Dict] = {},
        insight_id: Optional[str] = None,
    ):
        '''
        Remove document(s) and their embeddings from a vector database.

        Args:
            file_names (`List[str]`): List of files names that you want to remove from the db. The files should exist in the vector database.
            param_dict (`Optional[Dict]`): Additional parameters the engine might need to remove the documents.
            insight_id (`Optional[str]`): The insight ID to upload the documents to and process the request. Default is to use the clients current insight.

        Returns:
            `None`
        '''

        if insight_id == None:
            insight_id = self.insight_id

        param_dict['insight'] = insight_id

        return super().call(
            engine_type=VectorEngine.engine_type,
            engine_id=self.engine_id,
            method_name='removeDocument',
            method_args=[file_names, param_dict],
            method_arg_types=['java.util.List', 'java.util.Map'],
            insight_id=insight_id
        )

    def nearestNeighbor(
        self,
        search_statement: str,
        limit: Optional[int] = 5,
        param_dict: Optional[Dict] = {},
        insight_id: Optional[str] = None,
    ) -> List[Dict]:
        '''
        Perform a nearest neighbor or semantic search against a vector database. The searchStatement will be 
        converted to a vector using the same embedding model utilized to create the document(s) embeddings.


        Args:
            search_statement (`str`): The statement to search for semantic matches in the vector database
            limit (`Optional[int]`): The amount of top matches to return
            param_dict (`Optional[Dict]`): Additional parameters the engine might need to remove the documents.
            insight_id (`Optional[str]`): The insight ID to upload the documents to and process the request. Default is to use the clients current insight.

        Returns:
            `List[Dict]`: A list of dictionaries that contain the top semantic matches against the search statement
        '''

        if insight_id is None:
            if self.insight_id is None:
                insight_id = self.insight_id
            else:
                insight_id = self.server.cur_insight

        pixel = f"VectorDatabaseQuery(engine = \"{self.engine_id}\", command = [\"<e>{search_statement}</e>\"], limit = {limit}"

        if len(param_dict) != 0:
            pixel += ", paramValues = " + json.dumps(param_dict)

        pixel += ");"

        output_payload_message = self.server.run_pixel(
            payload=pixel,
            insight_id=insight_id,
            full_response=True
        )

        if (output_payload_message['pixelReturn'][0]['operationType'] == ['ERROR']):
            raise RuntimeError(
                output_payload_message['pixelReturn'][0]['output'])

        return output_payload_message['pixelReturn'][0]['output']

    def listDocuments(
        self,
        param_dict: Optional[Dict] = {},
        insight_id: Optional[str] = None,
    ) -> List[Dict]:
        '''
        List the set of documents a vector database is made from. At a minimum the file names should be returned but 
        the file size and last modified date are also options.

        Args:
            param_dict (`Optional[Dict]`): Additional parameters the engine might need to remove the documents.
            insight_id (`Optional[str]`): The insight ID to upload the documents to and process the request. Default is to use the clients current insight.

        Returns:
            `List[Dict]`: A list of dictionaries that contain file information
        '''
        if insight_id == None:
            insight_id = self.insight_id

        return super().call(
            engine_type=VectorEngine.engine_type,
            engine_id=self.engine_id,
            method_name='listDocuments',
            method_args=[param_dict],
            method_arg_types=['java.util.Map'],
            insight_id=insight_id
        )



from typing import Optional
import logging
from ...server_resources.server_proxy import ServerProxy

logger: logging.Logger = logging.getLogger(__name__)


class FunctionEngine(ServerProxy):

    def __init__(self, engine_id: str, insight_id: Optional[str] = None):
        super().__init__()

        self.engine_id = engine_id
        self.insight_id = insight_id

        logger.info("FunctionEngine initialized with engine id " + engine_id)

    def execute(
            self,
            parameterMap: dict,
            insight_id: Optional[str] = None
    ) -> None:
        """
        Connect to a function and execute

        Args:
            parameterMap (`dict`): A dictionary with the payload for the function engine
        """
        if insight_id is None:
            insight_id = self.insight_id

        pixel = f"ExecuteFunctionEngine(engine = \"{self.engine_id}\", map=[{parameterMap}]);"

        output_payload_message = self.server.run_pixel(
            payload=pixel,
            insight_id=insight_id,
            full_response=True
        )

        if (output_payload_message['pixelReturn'][0]['operationType'] == ['ERROR']):
            raise RuntimeError(
                output_payload_message['pixelReturn'][0]['output'])

        return output_payload_message['pixelReturn'][0]['output']



from typing import Optional
import logging
from ...server_resources.server_proxy import ServerProxy

logger: logging.Logger = logging.getLogger(__name__)


class StorageEngine(ServerProxy):

    def __init__(self, engine_id: str, insight_id: Optional[str] = None):
        super().__init__()

        self.engine_id = engine_id
        self.insight_id = insight_id

        logger.info("StorageEngine initialized with engine id " + engine_id)

    def list(self, path: str, insight_id: Optional[str] = None):

        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="list",
            method_args=[path],
            method_arg_types=["java.lang.String"],
        )

    def listDetails(self, path: str, insight_id: Optional[str] = None):
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="listDetails",
            method_args=[path],
            method_arg_types=["java.lang.String"],
        )

    def syncLocalToStorage(self, localPath: str, storagePath: str, insight_id: Optional[str] = None):
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="syncLocalToStorage",
            method_args=[localPath, storagePath],
            method_arg_types=["java.lang.String", "java.lang.String"],
        )

    def syncStorageToLocal(self, localPath: str, storagePath: str, insight_id: Optional[str] = None):
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="syncStorageToLocal",
            method_args=[storagePath, localPath],
            method_arg_types=["java.lang.String", "java.lang.String"],
        )

    def copyToLocal(self, storageFilePath: str, localFolderPath: str, insight_id: Optional[str] = None):
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="copyToLocal",
            method_args=[storageFilePath, localFolderPath],
            method_arg_types=["java.lang.String", "java.lang.String"],
        )

    def deleteFromStorage(self, storagePath, insight_id=None):
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="deleteFromStorage",
            method_args=[storagePath],
            method_arg_types=["java.lang.String"],
        )

    def deleteFromStorage(
        self, storagePath: str, leaveFolderStructure: Optional[bool] = False, insight_id: Optional[str] = None
    ):
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Storage",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="deleteFromStorage",
            method_args=[storagePath, leaveFolderStructure],
            method_arg_types=["java.lang.String", "java.lang.Boolean"],
        )



import unittest

from openai import OpenAI
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.completion import Completion
from openai.types.create_embedding_response import CreateEmbeddingResponse

from test_base_connection import TestServerClient
from variables import LLM_EMBEDDING_ENGINE_ID, LLM_CHAT_ENGINE_ID

class OpenAiEndpointsTests(TestServerClient):
    
    openai_client = None
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        if cls.openai_client is None:
            cls.openai_client = OpenAI(
                api_key="EMPTY",
                base_url=cls.server_client.get_openai_endpoint(),
                default_headers=cls.server_client.get_auth_headers()
            )
            
    def test_chat_completions_endpoint(self):
        response = self.openai_client.chat.completions.create(
            model=LLM_CHAT_ENGINE_ID, # change the model name to a Model Engine ID
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": "Who won the world series in 2020?"},
                {"role": "assistant", "content": "The Los Angeles Dodgers won the World Series in 2020."},
                {"role": "user", "content": "Where was it played?"}
            ]
        )
        
        self.assertIsInstance(response, ChatCompletion)
        self.assertIsNotNone(response.choices[0].message.content)        
        self.assertIsInstance(response.choices[0].message.content, str)
        
    def test_completions_endpoints(self):
        response = self.openai_client.completions.create(
            model=LLM_CHAT_ENGINE_ID,
            prompt="Write a tagline for an ice cream shop."
        )
        
        self.assertIsInstance(response, Completion)
        self.assertIsNotNone(response.choices[0].text)        
        self.assertIsInstance(response.choices[0].text, str)
        
    def test_embeddings_endpoints(self):
        response = self.openai_client.embeddings.create(
            model=LLM_EMBEDDING_ENGINE_ID,
            input=["Your text string goes here"]
        )
        
        self.assertIsInstance(response, CreateEmbeddingResponse)
        self.assertIsNotNone(response.data[0].embedding)        
        self.assertGreater(len(response.data[0].embedding), 0)
        

if __name__ == '__main__':
    unittest.main()



from typing import List, Optional, Dict, Union, Any, Generator
import logging
from ...server_resources.server_proxy import ServerProxy

logger: logging.Logger = logging.getLogger(__name__)


class ModelEngine(ServerProxy):

    def __init__(self, engine_id: Optional[str], insight_id: Optional[str] = None):
        super().__init__()
        self.engine_id = engine_id
        self.insight_id = insight_id

        logger.info("ModelEngine initialized with engine id " + engine_id)

    def ask(
        self,
        question: str,
        context: Optional[str] = None,
        insight_id: Optional[str] = None,
        param_dict: Optional[Dict] = None,
    ) -> List[Dict]:
        """
        Ask a question to a text-generation model hosted on the ai server.

        Args:
            question (`str`):
                The single string question you are asking a an LLM
            context (`str`):
                Set the given context string for an interaction with an LLM
            insight_id (`str`):
                Unique identifier for the temporal worksapce where actions are being isolated
            param_dict (`dict`):
                Additional inference params like temperature, max new tokens and top_k etc

                *NOTE* you can pass in
                param_dict = {"full_prompt":full_prompt, "temperature":temp, "max_new_tokens":max_token}
                where full_prompt is the an multi faceted prompt construct before sending the payload

                For OpenAI, this would be a list of dictionaris where the only keys within each dictionary are 'role' and 'content'
                For TextGen, this could be a list simialr to OpenAI or a complete string that has all the pieces pre constructed
        Response:
            `Union[Dict, Generator]`: A list that contains a dictioanry with the answer to the question, insight id and the message id
        """
        if insight_id is None:
            insight_id = self.insight_id

        full_response: List[Dict] = super().call(
            engine_type="Model",
            engine_id=self.engine_id,
            method_name="ask",
            method_args=[question, context, insight_id, param_dict],
            method_arg_types=[
                "java.lang.String",
                "java.lang.String",
                "prerna.om.Insight",
                "java.util.Map",
            ],
            insight_id=insight_id,
        )

        return full_response

    def stream_ask(
        self,
        question: str,
        context: Optional[str] = None,
        insight_id: Optional[str] = None,
        param_dict: Optional[Dict] = None,
    ) -> Generator:
        """
        Stream responses from a question asked to a text-generation model hosted on the ai server.

        Args:
            question (`str`):
                The single string question you are asking a an LLM
            context (`str`):
                Set the given context string for an interaction with an LLM
            insight_id (`str`):
                Unique identifier for the temporal worksapce where actions are being isolated
            param_dict (`dict`):
                Additional inference params like temperature, max new tokens and top_k etc

                *NOTE* you can pass in
                param_dict = {"full_prompt":full_prompt, "temperature":temp, "max_new_tokens":max_token}
                where full_prompt is the an multi faceted prompt construct before sending the payload

                For OpenAI, this would be a list of dictionaris where the only keys within each dictionary are 'role' and 'content'
                For TextGen, this could be a list simialr to OpenAI or a complete string that has all the pieces pre constructed
        Response:
            `Generator`: A generator to iterate through to get the partial responses
        """
        if insight_id is None:
            insight_id = self.insight_id

        for message in super().get_partial_responses(super().async_call(
            engine_type="Model",
            engine_id=self.engine_id,
            method_name="ask",
            method_args=[question, context, insight_id, param_dict],
            method_arg_types=[
                "java.lang.String",
                "java.lang.String",
                "prerna.om.Insight",
                "java.util.Map",
            ],
            insight_id=insight_id,
        )):
            yield message

    def embeddings(
        self,
        strings_to_embed: List[str],
        insight_id: Optional[str] = None,
        param_dict: Optional[Dict] = None,
    ) -> Dict:
        """
        Pass in a list of strings to an embeddings model and get a list of vectors.

        Args:
            question (`str`):
                A user's access key is a unique identifier used for authentication and authorization. It will allow users or applications to access specific resources or perform designated actions within an ai-server instance.
            insight_id (`str`):
                Unique identifier for the temporal worksapce where actions are being isolated
            param_dict (`dict`):
                Optional

        Returns

        """
        if isinstance(strings_to_embed, str):
            strings_to_embed = [strings_to_embed]
        assert isinstance(strings_to_embed, list)

        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Model",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="embeddings",
            method_args=[strings_to_embed, insight_id, param_dict],
            method_arg_types=["java.util.List",
                              "prerna.om.Insight", "java.util.Map"],
        )

    def model(
        self,
        input: Any,
        insight_id: Optional[str] = None,
        param_dict: Optional[Dict] = None,
    ) -> Any:
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Model",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="model",
            method_args=[input, insight_id, param_dict],
            method_arg_types=["java.lang.Object",
                              "prerna.om.Insight", "java.util.Map"],
        )

    def get_model_type(self, insight_id: Optional[str] = None) -> str:
        if insight_id is None:
            insight_id = self.insight_id

        return super().call(
            engine_type="Model",
            engine_id=self.engine_id,
            insight_id=insight_id,
            method_name="getModelType",
            method_args=[],
            method_arg_types=[],
        )

    def get_model_engine_id(self) -> str:
        return self.engine_id

    def to_langchain_embedder(self):
        """Transform the model engine into a langchain `Embeddings`object so that it can be used with langchain code"""

        from langchain_core.embeddings import Embeddings

        class CfgEmbeddingsEngine(Embeddings):
            def __init__(self, modelEngine):
                self.modelEngine = modelEngine

            def embed_documents(self, texts: List[str]) -> List[List[float]]:
                """Embed search docs."""
                return self.modelEngine.embeddings(strings_to_embed=texts)[0][
                    "response"
                ]

            def embed_query(self, text: str) -> List[float]:
                return self.modelEngine.embeddings(strings_to_embed=[text])[0][
                    "response"
                ][0]

        return CfgEmbeddingsEngine(modelEngine=self)

    def to_langchain_chat_model(self):
        """Transform the model engine into a langchain `BaseChatModel` object so that it can be used with langchain code"""
        from langchain_core.language_models.chat_models import BaseChatModel
        from langchain_core.outputs import (
            ChatGeneration,
            ChatResult,
        )
        from langchain_core.messages import (
            AIMessage,
            BaseMessage,
        )

        class ChatCfgAI(BaseChatModel):
            engine_id: str
            model_engine: ModelEngine
            model_type: str

            def __init__(self, model_engine):
                data = {
                    "engine_id": model_engine.get_model_engine_id(),
                    "model_engine": model_engine,
                    "model_type": model_engine.get_model_type()[0],
                }

                super().__init__(**data)

            class Config:
                """Configuration for this pydantic object."""

                allow_population_by_field_name = True

            def _generate(
                self,
                messages: List[BaseMessage],
                stop: Optional[List[str]] = None,
                **kwargs: Any,
            ) -> ChatResult:
                """Top Level call"""
                full_prompt = self.convert_messages_to_full_prompt(messages)
                response = self.model_engine.ask(
                    question="", param_dict={**kwargs, **{"full_prompt": full_prompt}}
                )

                return self._create_chat_result(response=response[0])

            def _create_chat_result(self, response: Dict[str, Any]) -> ChatResult:
                generations = []

                message = response.pop("response", "")
                generation_info = dict()
                if "logprobs" in response.keys():
                    generation_info["logprobs"] = response.pop("logprobs", {})
                gen = ChatGeneration(
                    message=AIMessage(content=message),
                    generation_info=generation_info,
                )

                generations.append(gen)

                return ChatResult(generations=generations, llm_output=response)

            def convert_messages_to_full_prompt(
                self,
                messages: List[BaseMessage],
            ) -> Union[Dict[str, Any], str]:
                """Convert a LangChain message to a the correct response for a model.

                Args:
                    message: The LangChain message.

                Returns:
                    The `Dict` or `str` containing the message payload.
                """

                if self.model_type in ["OPEN_AI", "VERTEX"]:
                    # assume this is a chat based openai model, otherwise why would you call this
                    # class
                    full_prompt: List[Dict[str, Any]]
                    from langchain_community.adapters.openai import (
                        convert_message_to_dict,
                    )

                    full_prompt = [
                        convert_message_to_dict(m) for m in messages]
                    return full_prompt
                else:
                    full_prompt: str
                    full_prompt = "\n".join([m.content for m in messages])
                    return full_prompt

            @property
            def _llm_type(self) -> str:
                """Return type of chat model."""
                return "CFG AI"

        return ChatCfgAI(model_engine=self)



from typing import (
    Any,
    List,
    Dict,
    Union,
    Optional,
    Set,
    Generator
)

import requests
import json
import pandas as pd
import base64
import logging

logger: logging.Logger = logging.getLogger(__name__)


class ServerClient():
    """ServerClient to make calls to a ai server instance

    Example:

    ```python
    >>> import ai_server

    # define access keys
    >>> loginKeys = {"secretKey":"<your_secret_key>","accessKey":"<your_access_key>"}

    # create connection object by passing in the secret key, access key and base url for the api
    >>> server_connection = ai_server.ServerClient(access_key=loginKeys['accessKey'], secret_key=loginKeys['secretKey'], base='<Your deployed server Monolith URL>')
    ```
    """

    # Class attribute to hold a singleton instance
    da_server = None

    def __init__(
        self,
        access_key: str,
        secret_key: str,
        base: str
    ) -> None:
        """
        Args:
            access_key (`str`):
                A user's access key is a unique identifier used for authentication and authorization. It will allow users or applications to access specific resources or perform designated actions within an ai-server instance.
            secret_key (`str`):
                A user's confidential and securely stored piece of information that complements an access key, used in cryptographic processes to validate identity and ensure secure communication
            base (`str`):
                main url to access the server api
        """
        self.access_key: str = access_key                             # set the access key as an instance attribute
        # set the secret key as an instance attribute
        self.secret_key: str = secret_key
        # set the base url as an instance attribute
        self.main_url: str = base

        if self.main_url.endswith('\/'):
            self.main_url = self.main_url[:-1]

        # TODO provide definitons for all of these attributes
        # used to keep track of the authorization header after the user has authenticated
        self.auth_headers: Dict = {}
        self.login()
        # boolean flag to make sure the users have authenticated and the cookies are set
        self.connected: bool = True
        # keep track of the the insights created from this connection
        self.open_insights: Set = set()
        self.cur_insight: str = self.make_new_insight()

        # map to keep track of requests and responses outside of this class or in separate threads
        self.monitors = {}

        # set the instance of this class as the class attribute da_server
        ServerClient.da_server = self

    def login(
        self
    ):
        """
        Register / validate the users secret & access key combination. 

        Store the cookies to be used for other api calls after authentication
        """
        combined = self.access_key + ":" + self.secret_key
        combined_enc = base64.b64encode(combined.encode('ascii'))
        headers = {
            'Authorization': f"Basic {combined_enc.decode('ascii')}"
        }
        self.auth_headers: Dict = headers.copy()
        logger.info(self.auth_headers)
        headers.update(
            {'disableRedirect': 'true'}
        )

        # authentication url
        api_url = "/auth/whoami"
        url = self.main_url + api_url

        self.r = requests.get(
            url,
            headers=headers
        )

        self.r.raise_for_status()

        # display the login response
        json_response = self.r.json()

        logger.info(json_response)

        if 'errorMessage' in json_response:
            if json_response['errorMessage'] == "null principal":
                raise AuthenticationError(
                    url + f" USERID = INVALID could not login using user access key '{self.access_key}' with invalid secret key")

            raise AuthenticationError(json_response['errorMessage'])

        self.cookies = self.r.cookies

        # display the cookies
        logger.info(self.cookies)

    # TODO add dec
    def get_auth_headers(
        self
    ) -> Dict:
        '''Get the autheroization headers used to authenticate the user'''
        return self.auth_headers

    def get_openai_endpoint(
        self
    ):
        '''Get the corrensponding openai endpoint for the CFG AI server'''
        return self.main_url + "/model/openai"

    def make_new_insight(
        self
    ) -> str:
        """
        Create a new insight (temporal space) to operate within the ai-server at set it as the current insight.
        """
        if not self.connected:
            return "Please login"

        response = requests.post(
            self.main_url + "/engine/runPixel",
            cookies=self.cookies,
            data={
                'expression': 'META | true',
                'insightId': 'new'
            }
        )

        # raise HTTP error if one occurs
        response.raise_for_status()

        json_output = response.json()
        logger.info(json_output)
        self.cur_insight = json_output['insightID']
        self.open_insights.add(self.cur_insight)

        logger.info(f"Current open insights are -- {self.open_insights}")

        return self.cur_insight

    def run_pixel(
        self,
        payload: str,
        insight_id: Optional[str] = None,
        full_response: Optional[bool] = False
    ):
        '''
        Send a pixel payload to the platforms /api/engine/runPixel endpoint.

        /api/engine/runPixel is the AI server's primary endpoint that consumes a flexible payload. The payload must contain two parameters, namely:
            1.) expression - The @payload passed is placed here and must comply and be defined in the Server's DSL (Pixel) which dictates what action should be taken

            2.) insightId - the temporal workspace identifier which isoloates the actions being performed

        Args:
            payload (`str`): DSL (Pixel) instruction on what specific action should be performed 
            insight_id (`str`): Unique identifier for the temporal worksapce where actions are being isolated. Options are:
                - insight_id = '' -> Creates a temporary one time insight created but is not stored /     cannot be referenced for future state
                - insight_id = 'new' -> Creates a new insight which can be referenced in the same sesssion
                - insight_id = '<uuid/guid string>' -> Uses an existing insight id that exists in the user session
            full_response (`bool`): Indicate whether to return the full json response or only the actions output

        Returns:
            `Union[Any, Dict]`: The output object from the runPixel response or the entire runPixel response.
        '''
        # going to create an insight if insight not available
        if not self.connected:
            return "Please login"

        if insight_id is None:
            insight_id = self.cur_insight
            if insight_id is None:
                # the insight_id is still null
                logger.info(
                    "insight_id and self.cur_insight are both undefined. Creating new insight")
                self.cur_insight = self.make_new_insight()
                insight_id = self.cur_insight

        logger.info(f"Current insight_id is set to {insight_id}")

        pixel_payload = {
            'expression': payload,
            'insightId': insight_id
        }

        api_url = "/engine/runPixel"
        response = requests.post(
            self.main_url + api_url,
            cookies=self.cookies,
            data=pixel_payload
        )

        response_dict = response.json()
        logger.info(response_dict)
        if ("ERROR" in response_dict['pixelReturn'][0]['operationType']):
            raise Exception(response_dict['pixelReturn'][0]['output'])

        if full_response:
            return response_dict
        else:
            return self.get_pixel_output(response_dict)

    def get_partial_responses(self, job_id: str) -> Generator:
        '''
        Make continuois post requests to the partial endpoint until we see a change in the message state

        Args:
            job_id (`str`): The identifier of the job being run

        Returns:
            `Generator`: A generator that can be iterated through to get partials
        '''

        assert job_id is not None, "Please provide a valid job ID."

        partial_endpoint = self.main_url + '/engine/partial'
        cookies = self.cookies
        payload = {
            'jobId': job_id,
        }

        # keep track of when the streams start coming since it could be delayed
        started_streaming = False
        while True:
            response = requests.post(
                partial_endpoint, cookies=cookies, data=payload).json()
            msg = response.get('message')
            msg_length = len(msg)

            if msg_length != 0 and not started_streaming:
                started_streaming = True
                yield msg['new']
            elif msg_length != 0:
                yield msg['new']
            elif msg_length == 0 and started_streaming:
                break

    def get_pixel_output(
        self,
        response: Dict
    ) -> Union[Any, List]:
        '''
        Utility method to grab the output of runPixel call.

        Args:
            response (`Dict`): The runPixel response from the Tomcat Server

        Returns:
            `Union[Any, List]`: The output object or a list of objects 
        '''
        main_output = response['pixelReturn'][0]['output']

        if isinstance(main_output, list):
            output = main_output[0]
        else:
            output = main_output

        return output

    def logout(
        self
    ) -> None:
        '''Closes the connection to the server.'''
        requests.get(self.main_url + "/logout/all", cookies=self.cookies)

        self.cookies = None         # reset the connection attributes for the class
        self.connected = False      # reset the connection attributes for the class

    def send_request(
        self,
        payload_struct: Dict
    ) -> None:
        '''
        Constructs the pixel payload from a PayloadStruct for various server resources such as ModelEngine, StorageEngine and DatabaseEngine

        Args:
            payload_struct (`dict`): the actual payload being sent to the AI Server

        Returns:
            `None`
        '''

        # this is a synchronous call
        # but I dont want to bother the server proxy and leave it as is
        epoc = payload_struct['epoc']

        input_payload_message = json.dumps(payload_struct)

        logger.info("Sending a PayloadStruct " + input_payload_message)

        # RemoteEngineRun is responsible for handling ModelEngine, StorageEngine and DatabaseEngine via RESTServer
        output_payload_message = self.run_pixel(
            payload="RemoteEngineRun(payload=\"<e>" +
            input_payload_message + "</e>\");",
            insight_id=payload_struct['insightId']
        )

        if epoc in self.monitors:
            self.monitors[epoc] = output_payload_message

    def get_open_insights(
        self
    ) -> List[str]:
        '''
        Retrieves a list of insight IDs created after establishing the connection.

        Args:

        Returns:
            `List[str]`: List of insight IDs
        '''
        # MyOpenInsights is pre0defined server level reactor to handle the action of getting the insightIds
        open_insights = self.run_pixel(payload='MyOpenInsights();')

        # keep track of the open insights within the python object itself
        self.open_insights = set(open_insights)

        # return a list of insight ids
        return open_insights

    def drop_insights(
        self,
        insight_ids: Union[str, List[str]]
    ) -> None:
        '''
        Utility method to close insight(s). By default, if the the current/working insight is closed, then we will attempt to set to the first insight in open_insights as the current insight

        Args:
            insight_ids (`str` | `list`): a single insight or a list of insights the user wishes to close

        Returns:
            `None`
        '''
        if isinstance(insight_ids, list):
            for id in insight_ids:
                self.run_pixel(insight_id=id, payload='DropInsight();')
                self.open_insights.discard(id)
        else:
            self.run_pixel(insight_id=insight_ids, payload='DropInsight();')
            self.open_insights.discard(insight_ids)

        if self.cur_insight not in self.open_insights:
            if (len(self.open_insights) > 0):
                self.cur_insight = self.open_insights[0]
            else:
                self.cur_insight = None

    def import_data_product(
        self,
        project_id: str,
        insight_id: str,
        sql: str
    ) -> Union[pd.DataFrame, None]:
        '''
        Accesses data within a saved insight (data product) via a REST API.

        Args:
        project_id (`str`):
            Given project/app unique identifier 
        insight_id (`str`):
            Shared insight indentifier.
        sql (`bool`):
            SQL statement that is executable on the frames within the data product

        Returns:
          `pd.DataFrame` Pandas Dataframe based on the SQL statement passed in
        '''
        base_url = self.main_url + '/project-' + project_id + \
            '/jdbc_json?insightId=' + insight_id + '&open=true&sql='

        dataProductUrl = base_url + sql
        response = requests.get(
            dataProductUrl, cookies=self.cookies
        ).json()

        try:
            return pd.DataFrame(response['dataArray'], columns=response['columns'])
        except:
            try:
                return pd.DataFrame(response['data'], columns=response['columns'])
            except:
                return None

    def upload_files(
        self,
        files: List[str],
        project_id: Optional[str] = None,
        insight_id: Optional[str] = None,
        path: Optional[str] = None
    ) -> List[str]:
        '''
        Uploads files from the local device to the server.

        Args:
            files (`List[str]`):
                List of file paths to upload from local device
            insight_id (`str`):
                Unique identifier for the temporal worksapce where actions are being isolated
            project_id (`bool`):
                Given project/app unique identifier
            path (`str`):
                Specific upload path

        Returns (`List[str]`):
            List of file names that have been successfully uploaded

        '''
        # .../Monolith/api/uploadFile/baseUpload?insightId=de43ce0d-db2e-4ab9-a807-336bb86c4ea0&projectId=4c14bc58-973f-4293-87ed-a5d32c24f418&path=version/assets/

        if isinstance(files, str):
            files = [files]

        param = ''
        path = path or ''

        if insight_id or project_id or path:
            if insight_id == None:
                insight_id = self.cur_insight

            param += f'insightId={insight_id}'

            if project_id:
                if param:
                    param += '&'
                param += f'projectId={project_id}'

            if path:
                if param:
                    param += '&'
                param += f'path={path}'
        else:
            param += f'insightId={self.cur_insight}'

        param = f'?{param}'

        url = f'{self.main_url}/uploadFile/baseUpload{param}'

        logger.info("The upload url is " + url)

        insight_file_paths = []
        for filepath in files:
            with open(filepath, 'rb') as fobj:
                response = requests.post(
                    url,
                    cookies=self.cookies,
                    files={'file': fobj}
                )
                insight_file_paths.append(response.json()[0]['fileName'])

        return insight_file_paths


class AuthenticationError(Exception):
    pass
